// /Applications/MATLAB2009a/MATLAB2009a.app/bin/mex

#include "mex.h"
#include "matrix.h"
#include "ray_main_mex.h"

extern void _main();

/*
 INPUT:
 Vxz - matrix[nx][nz] of uniformly sampled velocities v(x,z)
 X0  - x coordinate of takeoff point
 Z0  - z coordinate of takeoff point
 A0  - z coordinate of takeoff point
 Dt  - time sampling interval
 Ft  - first time sample
 Dx  - x sampling interval
 Fx  - first x sample
 Dz  - z sampling interval
 Fz  - first z sample
 Nx  - number of x samples
 Nz  - number of z samples
 Nt  - number of time samples
 
 OUTPUT: 
 matrix out[nt][n_var]
 1  x   - x position
 2  z   - z position
 3  t   - time
 4  a   - angle
 5  c   - cosine
 6  s   - sine
 7  q1  - dynamic ray tracing solutions
 8  p1  
 9  q2   
 10 p2   
 11 kmah- kmah index
 12 v   - velocity along ray
 
 */

void mexFunction(int nlhs, mxArray* plhs[], int nrhs, const mxArray* prhs[])
{
	if(nrhs<13)
		mexErrMsgTxt("at least 13 inputs required");
	//Declarations
	// INPUT
	mxArray *Vxz;		// velocity
	mxArray *X0, *Z0, *A0, *Dt, *Ft, *Dx, *Fx, *Dz, *Fz;
	mxArray *Nx, *Nt, *Nz, *cflag;
	// OUTPUT
	double *xray;  // x coord
	double *zray;  // z coord
	// INTERN
	double *vxz, *xx0, *zz0, *aa0, *ddt;
	float **xzray;
	float **vXZ;
	double *xzarray;
	float *x0, *z0, *a0, *dt, ft, dx, fx, dz, fz;
	int nx, nt, nz, it, itx, itz,dimr, siz1, CFLAG;
	
	dimr=12; // number of output variables
	
	//Copy input pointers
	Vxz= prhs[0]; 
	X0 = prhs[1];
	Z0 = prhs[2];
	A0 = prhs[3];
	Dt = prhs[4];
	Ft = prhs[5];
	Dx = prhs[6];
	Fx = prhs[7];
	Dz = prhs[8];
	Fz = prhs[9];
	Nx = prhs[10];
	Nz = prhs[11];
	Nt = prhs[12];
	if(nrhs>13) cflag = prhs[13];
		
	
	//Get input
	vxz = mxGetPr(Vxz);
	/*	x0 = (float)(mxGetScalar(X0));
	 z0 = (float)(mxGetScalar(Z0));
	 a0 = (float)(mxGetScalar(A0));
	 */
	xx0 = (mxGetPr(X0));
	zz0 = (mxGetPr(Z0));
	aa0 = (mxGetPr(A0));
	siz1 = mxGetM(X0);	// legendre max order
//	dt = (float)(mxGetScalar(Dt));
	ddt = (mxGetPr(Dt));
	ft = (float)(mxGetScalar(Ft));
	dx = (float)(mxGetScalar(Dx));
	fx = (float)(mxGetScalar(Fx));
	dz = (float)(mxGetScalar(Dz));
	fz = (float)(mxGetScalar(Fz));
	nx = (int)(mxGetScalar(Nx));
	nz = (int)(mxGetScalar(Nz));
	nt = (int)(mxGetScalar(Nt));
	if(nrhs>13)
	{
		CFLAG=(int)(mxGetScalar(cflag));
	}
	else
	{
		CFLAG=0;
	}
	
	x0=(float*)malloc(siz1*sizeof(float*));z0=(float*)malloc(siz1*sizeof(float*));a0=(float*)malloc(siz1*sizeof(float*));dt=(float*)malloc(siz1*sizeof(float*));
	
	for (itx=0; itx<siz1; itx++) {
		x0[itx]=(float)xx0[itx];z0[itx]=(float)zz0[itx];a0[itx]=(float)aa0[itx];dt[itx]=(float)ddt[itx];
	}
	
	
	vXZ=(float**)malloc(nx*sizeof(float*)); // allocate v
	for (itx=0; itx<nx; itx++) {
		vXZ[itx]=(float*)malloc(nz*sizeof(float));
	}
	for (itx=0; itx<nx; ++itx) {			// read v
		for (itz=0; itz<nz; ++itz) {
			vXZ[itx][itz]=(float)vxz[itx+itz*nx];
		}
	}
	
	
	//Allocate memory and assign output pointer
	plhs[0] = mxCreateDoubleMatrix(siz1, dimr,  mxREAL); //mxReal is our data-type
	xzarray=mxGetPr(plhs[0]);
	
	xzray=(float**)malloc(dimr*sizeof(float*));
	for (it=0; it<dimr; it++) {
		xzray[it]=(float*)malloc(siz1*sizeof(float));
	}
	
//	printf("siz1=%d\n",siz1);
	
	makeRayField ( x0, z0, a0,  nt, dt, ft, nx, dx, fx, nz, dz, fz, vXZ, xzray, siz1, CFLAG);
	
	for (it=0; it<dimr; it++) {
		for (itx=0; itx<siz1; itx++) {
			xzarray[itx+siz1*it]=(double)xzray[it][itx];
		}
	}
	
	// free matrices
	for (itx=0; itx<nx; itx++)		
		free(vXZ[itx]);		
	free(vXZ);
	for (it=0; it<dimr; it++) {
		free(xzray[it]);
	}
	free(xzray);
	free(x0);free(z0);free(a0);
	
	return;	
}


void **makeRayField (float *x0, float *z0, float *a0, int nt, float *DT, float ft,
					 int nx, float dx, float fx, int nz, float dz, float fz, float **vxz, float **xzray, int sizpos, int CFLAG)
/*****************************************************************************
 Trace a ray for uniformly sampled v(x,z).
 ******************************************************************************
 Input:
 x0		x coordinate of takeoff point
 z0		z coordinate of takeoff point
 a0		takeoff angle (radians)
 nt		number of time samples
 dt		time sampling interval
 ft		first time sample
 nx		number of x samples
 dx		x sampling interval
 fx		first x sample
 nz		number of z samples
 dz		z sampling interval
 fz		first z sample
 vxz		array[nx][nz] of uniformly sampled velocities v(x,z)
 
 Returned:	pointer to ray parameters sampled at discrete ray steps
 ******************************************************************************
 Notes:
 The ray ends when it runs out of time (after nt steps) or with the first 
 step that is out of the (x,z) bounds of the velocity function v(x,z).
 *****************************************************************************/
{
	int it,kmah,itt;
	float t,x,z,a,c,s,p1,q1,p2,q2,
	lx,lz,cc,ss,
	v,dvdx,dvdz,ddvdxdx,ddvdxdz,ddvdzdz,
	vv,ddvdndn,
	dt;
	void *vel2;
	float xp, zp, tp, ap, cp, sp, q1p, p1p, q2p, p2p, kmahp, vp;
	
	/* allocate and initialize velocity interpolator */
	vel2 = vel2Alloc(nx,dx,fx,nz,dz,fz,vxz);
	
	/* last x and z in velocity model */
	lx = fx+(nx-1)*dx;
	lz = fz+(nz-1)*dz;
	
	/* allocate space for ray endpoints */
	for (it=0; it<12; it++) {
		for (itt=0; itt<sizpos; itt++) {
			xzray[it][itt]=(double)0.0;
		}
	}
	
	// loop positions
	for (itt=0; itt<sizpos; itt++) {
		dt=DT[itt];
		/* ensure takeoff point is within model */	
		if (x0[itt]<fx || x0[itt]>lx || z0[itt]<fz || z0[itt]>lz)
		{
			xzray[0][itt]=(double)0.0;
			xzray[1][itt]=(double)0.0;
			xzray[2][itt]=(double)0.0;
			xzray[3][itt]=(double)0.0;
			xzray[4][itt]=(double)0.0;
			xzray[5][itt]=(double)0.0;
			xzray[6][itt]=(double)0.0;
			xzray[7][itt]=(double)0.0;
			xzray[8][itt]=(double)0.0;
			xzray[9][itt]=(double)0.0;
			xzray[10][itt]=(double)0.0;
			xzray[11][itt]=(double)0.0;
		}
		else
		{
			/* cosine and sine of takeoff angle */
			c = cos(a0[itt]);
			s = sin(a0[itt]);
			cc = c*c;
			ss = s*s;
			
			/* velocity and derivatives at takeoff point */
			vel2Interp(vel2,x0[itt],z0[itt],&v,&dvdx,&dvdz,&ddvdxdx,&ddvdxdz,&ddvdzdz);
			ddvdndn = ddvdxdx*cc-2.0*ddvdxdz*s*c+ddvdzdz*ss;
			vv = v*v;
			
			/* first ray step */
			t = ft;a = a0[itt];x = x0[itt];z = z0[itt];q1 = 1.0;p1 = 0.0;q2 = 0.0;p2 = 1.0;kmah = 0;
			
//			printf("ray %d start: x=%2.1f, z=%2.1f, t=%2.1f\n",itt,x,z,t);
			
			/* loop over time steps */
			for (it=1; it<nt; ++it) {
				// keep old values
				xp=x; zp=z; tp=t; ap=a; cp=c; sp=s; q1p=q1; p1p=p1; q2p=q2; p2p=p2; kmahp=kmah; vp=v;
				/* variables used for Runge-Kutta integration */
				float h=dt,hhalf=dt/2.0,hsixth=dt/6.0,
				q2old,xt,zt,at,p1t,q1t,p2t,q2t,
				dx,dz,da,dp1,dq1,dp2,dq2,
				dxt,dzt,dat,dp1t,dq1t,dp2t,dq2t,
				dxm,dzm,dam,dp1m,dq1m,dp2m,dq2m;
				
				/* if ray is out of bounds, break */
				if (x<fx || x>lx || z<fz || z>lz)
				{ 
//					printf("ray %d: out of bound at iteration %d ...\n",itt, it);
					x=xp; z=zp; t=tp; a=ap; c=cp; s=sp; q1=q1p; p1=p1p; q2=q2p; p2=p2p; kmah=kmahp; v=vp; break;
				}
				
				/* remember old q2 */
//				q2old = q2;
				q2old = q1;
				
				/* step 1 of 4th-order Runge-Kutta */
				dx =  v*s;
				dz =  v*c;
				da =  dvdz*s-dvdx*c;
				dp1 = -ddvdndn*q1/v;
				dq1 = vv*p1;
				dp2 = -ddvdndn*q2/v;
				dq2 =  vv*p2;
				xt = x+hhalf*dx;
				zt = z+hhalf*dz;
				at = a+hhalf*da;
				p1t = p1+hhalf*dp1;
				q1t = q1+hhalf*dq1;
				p2t = p2+hhalf*dp2;
				q2t = q2+hhalf*dq2;
				c = cos(at);
				s = sin(at);
				cc = c*c;
				ss = s*s;
				vel2Interp(vel2,xt,zt,
						   &v,&dvdx,&dvdz,&ddvdxdx,&ddvdxdz,&ddvdzdz);
				ddvdndn = ddvdxdx*cc-2.0*ddvdxdz*s*c+ddvdzdz*ss;
				vv = v*v;
				
				/* step 2 of 4th-order Runge-Kutta */
				dxt =  v*s;
				dzt =  v*c;
				dat =  dvdz*s-dvdx*c;
				dp1t = -ddvdndn*q1t/v;
				dq1t = vv*p1t;
				dp2t = -ddvdndn*q2t/v;
				dq2t =  vv*p2t;
				xt = x+hhalf*dxt;
				zt = z+hhalf*dzt;
				at = a+hhalf*dat;
				p1t = p1+hhalf*dp1t;
				q1t = q1+hhalf*dq1t;
				p2t = p2+hhalf*dp2t;
				q2t = q2+hhalf*dq2t;
				c = cos(at);
				s = sin(at);
				cc = c*c;
				ss = s*s;
				vel2Interp(vel2,xt,zt,
						   &v,&dvdx,&dvdz,&ddvdxdx,&ddvdxdz,&ddvdzdz);
				ddvdndn = ddvdxdx*cc-2.0*ddvdxdz*s*c+ddvdzdz*ss;
				vv = v*v;
				
				/* step 3 of 4th-order Runge-Kutta */
				dxm =  v*s;
				dzm =  v*c;
				dam =  dvdz*s-dvdx*c;
				dp1m = -ddvdndn*q1t/v;
				dq1m = vv*p1t;
				dp2m = -ddvdndn*q2t/v;
				dq2m =  vv*p2t;
				xt = x+h*dxm;
				zt = z+h*dzm;
				at = a+h*dam;
				p1t = p1+h*dp1m;
				q1t = q1+h*dq1m;
				p2t = p2+h*dp2m;
				q2t = q2+h*dq2m;
				dxm += dxt;
				dzm += dzt;
				dam += dat;
				dp1m += dp1t;
				dq1m += dq1t;
				dp2m += dp2t;
				dq2m += dq2t;
				c = cos(at);
				s = sin(at);
				cc = c*c;
				ss = s*s;
				vel2Interp(vel2,xt,zt,
						   &v,&dvdx,&dvdz,&ddvdxdx,&ddvdxdz,&ddvdzdz);
				ddvdndn = ddvdxdx*cc-2.0*ddvdxdz*s*c+ddvdzdz*ss;
				vv = v*v;
				
				/* step 4 of 4th-order Runge-Kutta */
				dxt =  v*s;
				dzt =  v*c;
				dat =  dvdz*s-dvdx*c;
				dp1t = -ddvdndn*q1t/v;
				dq1t = vv*p1t;
				dp2t = -ddvdndn*q2t/v;
				dq2t =  vv*p2t;
				x += hsixth*(dx+dxt+2.0*dxm);
				z += hsixth*(dz+dzt+2.0*dzm);
				a += hsixth*(da+dat+2.0*dam);
				p1 += hsixth*(dp1+dp1t+2.0*dp1m);
				q1 += hsixth*(dq1+dq1t+2.0*dq1m);
				p2 += hsixth*(dp2+dp2t+2.0*dp2m);
				q2 += hsixth*(dq2+dq2t+2.0*dq2m);
				c = cos(a);
				s = sin(a);
				cc = c*c;
				ss = s*s;
				vel2Interp(vel2,x,z,
						   &v,&dvdx,&dvdz,&ddvdxdx,&ddvdxdz,&ddvdzdz);
				ddvdndn = ddvdxdx*cc-2.0*ddvdxdz*s*c+ddvdzdz*ss;
				vv = v*v;
				
				/* update kmah index */
//				if ((q2<=0.0 && q2old>0.0) || (q2>=0.0 && q2old<0.0))
				if ((q1<=0.0 && q2old>0.0) || (q1>=0.0 && q2old<0.0))
				{
					kmah++;
					if (CFLAG!=0)
					{
//						printf("kmah=%d (ray %d, it %d)\n",kmah,itt,it);
//						x=xp; z=zp; 
//						q2=q2p;
						break;
					}
				}
				
				/* update time */
				t += dt;
			}
			/* free velocity interpolator */
			/* save ray parameters */
//			printf("ray %d, it %d: kmah=%d, CFLAG=%d\n",itt,it,kmah,CFLAG);
			xzray[0][itt]=(double)x;
			xzray[1][itt]=(double)z;
			xzray[2][itt]=(double)t;
			xzray[3][itt]=(double)a;
			xzray[4][itt]=(double)c;
			xzray[5][itt]=(double)s;
			xzray[6][itt]=(double)q1;
			xzray[7][itt]=(double)p1;
			xzray[8][itt]=(double)q2;
			xzray[9][itt]=(double)p2;
			xzray[10][itt]=(double)kmah;
			xzray[11][itt]=(double)v;
			
//			printf("        stop: x=%2.1f, z=%2.1f, t=%2.1f, it=%d\n",x,z,t,it);
		}	
	}
	vel2Free(vel2);		
	return;
}

/*****************************************************************************
 Functions to support interpolation of velocity and its derivatives.
 ******************************************************************************
 Functions:
 vel2Alloc	allocate and initialize an interpolator for v(x,z)
 vel2Interp	interpolate v(x,z) and its derivatives
 ******************************************************************************
 Notes:
 Interpolation is performed by piecewise cubic Hermite polynomials,
 so that velocity and first derivatives are continuous.  Therefore,
 velocity v, first derivatives dv/dx and dv/dz, and the mixed
 derivative ddv/dxdz are continuous.  However, second derivatives
 ddv/dxdx and ddv/dzdz are discontinuous.
 *****************************************************************************/

/* number of pre-computed, tabulated interpolators */
#define NTABLE 101

/* length of each interpolator in table (4 for piecewise cubic) */
#define LTABLE 4

/* table of pre-computed interpolators, for 0th, 1st, and 2nd derivatives */
static float tbl[3][NTABLE][LTABLE];

/* constants */
static int ix=1-LTABLE/2-LTABLE,iz=1-LTABLE/2-LTABLE;
static float ltable=LTABLE,ntblm1=NTABLE-1;

/* indices for 0th, 1st, and 2nd derivatives */
static int kx[6]={0,1,0,2,1,0};
static int kz[6]={0,0,1,0,1,2};

/* function to build interpolator tables; sets tabled=1 when built */
static void buildTables (void);
static int tabled=0;

/* interpolator for velocity function v(x,z) of two variables */
typedef struct Vel2Struct {
	int nx;		/* number of x samples */
	int nz;		/* number of z samples */
	int nxm;	/* number of x samples minus LTABLE */
	int nzm;	/* number of x samples minus LTABLE */
	float xs,xb,zs,zb,sx[3],sz[3],**vu;
} Vel2;

void* vel2Alloc (int nx, float dx, float fx,
				 int nz, float dz, float fz, float **v)
/*****************************************************************************
 Allocate and initialize an interpolator for v(x,z) and its derivatives.
 ******************************************************************************
 Input:
 nx		number of x samples
 dx		x sampling interval
 fx		first x sample
 nz		number of z samples
 dz		z sampling interval
 fz		first z sample
 v		array[nx][nz] of uniformly sampled v(x,z)
 
 Returned:	pointer to interpolator
 *****************************************************************************/
{
	Vel2 *vel2;
	
	/* allocate space */
	/* HW */
	/*	vel2 = (Vel2*)alloc1(1,sizeof(Vel2));*/
	vel2 = (Vel2*)malloc(1*sizeof(Vel2));
	
	/* set state variables used for interpolation */
	vel2->nx = nx;
	vel2->nxm = nx-LTABLE;
	vel2->xs = 1.0/dx;
	vel2->xb = ltable-fx*vel2->xs;
	vel2->sx[0] = 1.0;
	vel2->sx[1] = vel2->xs;
	vel2->sx[2] = vel2->xs*vel2->xs;
	vel2->nz = nz;
	vel2->nzm = nz-LTABLE;
	vel2->zs = 1.0/dz;
	vel2->zb = ltable-fz*vel2->zs;
	vel2->sz[0] = 1.0;
	vel2->sz[1] = vel2->zs;
	vel2->sz[2] = vel2->zs*vel2->zs;
	vel2->vu = v;
	
	/* if necessary, build interpolator coefficient tables */
	if (!tabled) buildTables();
	
	return vel2;
}

void vel2Free (void *vel2)
/*****************************************************************************
 Free an interpolator for v(x,z) and its derivatives.
 ******************************************************************************
 Input:
 vel2		pointer to interpolator as returned by vel2Alloc()
 *****************************************************************************/
{
	free(vel2);
}


void vel2Interp (void *vel2, float x, float z,
				 float *v, float *vx, float *vz, float *vxx, float *vxz, float *vzz)
/*****************************************************************************
 Interpolation of a velocity function v(x,z) and its derivatives.
 ******************************************************************************
 Input:
 vel2		pointer to interpolator as returned by vel2Alloc()
 x		x coordinate at which to interpolate v(x,z) and derivatives
 z		z coordinate at which to interpolate v(x,z) and derivatives
 
 Output:
 v		v(x,z)
 vx		dv/dx
 vz		dv/dz
 vxx		ddv/dxdx
 vxz		ddv/dxdz
 vzz		ddv/dzdz
 *****************************************************************************/
{
	Vel2 *v2=vel2;
	int nx=v2->nx,nz=v2->nz,nxm=v2->nxm,nzm=v2->nzm;
	float xs=v2->xs,xb=v2->xb,zs=v2->zs,zb=v2->zb,
	*sx=v2->sx,*sz=v2->sz,**vu=v2->vu;
	int i,jx,lx,mx,jz,lz,mz,jmx,jmz,mmx,mmz;
	float ax,bx,*px,az,bz,*pz,sum,vd[6];
	
	/* determine offsets into vu and interpolation coefficients */
	ax = xb+x*xs;
	jx = (int)ax;
	bx = ax-jx;
	lx = (bx>=0.0)?bx*ntblm1+0.5:(bx+1.0)*ntblm1-0.5;
	lx *= LTABLE;
	mx = ix+jx;
	az = zb+z*zs;
	jz = (int)az;
	bz = az-jz;
	lz = (bz>=0.0)?bz*ntblm1+0.5:(bz+1.0)*ntblm1-0.5;
	lz *= LTABLE;
	mz = iz+jz;
	
	/* if totally within input array, use fast method */
	if (mx>=0 && mx<=nxm && mz>=0 && mz<=nzm) {
		for (i=0; i<6; ++i) {
			px = &(tbl[kx[i]][0][0])+lx;
			pz = &(tbl[kz[i]][0][0])+lz;
			vd[i] = sx[kx[i]]*sz[kz[i]]*(
										 vu[mx][mz]*px[0]*pz[0]+
										 vu[mx][mz+1]*px[0]*pz[1]+
										 vu[mx][mz+2]*px[0]*pz[2]+
										 vu[mx][mz+3]*px[0]*pz[3]+
										 vu[mx+1][mz]*px[1]*pz[0]+
										 vu[mx+1][mz+1]*px[1]*pz[1]+
										 vu[mx+1][mz+2]*px[1]*pz[2]+
										 vu[mx+1][mz+3]*px[1]*pz[3]+
										 vu[mx+2][mz]*px[2]*pz[0]+
										 vu[mx+2][mz+1]*px[2]*pz[1]+
										 vu[mx+2][mz+2]*px[2]*pz[2]+
										 vu[mx+2][mz+3]*px[2]*pz[3]+
										 vu[mx+3][mz]*px[3]*pz[0]+
										 vu[mx+3][mz+1]*px[3]*pz[1]+
										 vu[mx+3][mz+2]*px[3]*pz[2]+
										 vu[mx+3][mz+3]*px[3]*pz[3]);
		}
		
		/* else handle end effects with constant extrapolation */
	} else {
		for (i=0; i<6; ++i) {
			px = &(tbl[kx[i]][0][0])+lx;
			pz = &(tbl[kz[i]][0][0])+lz;
			for (jx=0,jmx=mx,sum=0.0; jx<4; ++jx,++jmx) {
				mmx = jmx;
				if (mmx<0) mmx = 0;
				else if (mmx>=nx) mmx = nx-1;
				for (jz=0,jmz=mz; jz<4; ++jz,++jmz) {
					mmz = jmz;
					if (mmz<0) mmz = 0;
					else if (mmz>=nz) mmz = nz-1;
					sum += vu[mmx][mmz]*px[jx]*pz[jz];
				}
			}
			vd[i] = sx[kx[i]]*sz[kz[i]]*sum;
		}
	}
	
	/* set output variables */
	*v = vd[0];
	*vx = vd[1];
	*vz = vd[2];
	*vxx = vd[3];
	*vxz = vd[4];
	*vzz = vd[5];
}

/* hermite polynomials */
static float h00 (float x) {return 2.0*x*x*x-3.0*x*x+1.0;}
static float h01 (float x) {return 6.0*x*x-6.0*x;}
static float h02 (float x) {return 12.0*x-6.0;}
static float h10 (float x) {return -2.0*x*x*x+3.0*x*x;}
static float h11 (float x) {return -6.0*x*x+6.0*x;}
static float h12 (float x) {return -12.0*x+6.0;}
static float k00 (float x) {return x*x*x-2.0*x*x+x;}
static float k01 (float x) {return 3.0*x*x-4.0*x+1.0;}
static float k02 (float x) {return 6.0*x-4.0;}
static float k10 (float x) {return x*x*x-x*x;}
static float k11 (float x) {return 3.0*x*x-2.0*x;}
static float k12 (float x) {return 6.0*x-2.0;}

/* function to build interpolation tables */
static void buildTables(void)
{
	int i;
	float x;
	
	/* tabulate interpolator for 0th derivative */
	for (i=0; i<NTABLE; ++i) {
		x = (float)i/(NTABLE-1.0);
		tbl[0][i][0] = -0.5*k00(x);
		tbl[0][i][1] = h00(x)-0.5*k10(x);
		tbl[0][i][2] = h10(x)+0.5*k00(x);
		tbl[0][i][3] = 0.5*k10(x);
		tbl[1][i][0] = -0.5*k01(x);
		tbl[1][i][1] = h01(x)-0.5*k11(x);
		tbl[1][i][2] = h11(x)+0.5*k01(x);
		tbl[1][i][3] = 0.5*k11(x);
		tbl[2][i][0] = -0.5*k02(x);
		tbl[2][i][1] = h02(x)-0.5*k12(x);
		tbl[2][i][2] = h12(x)+0.5*k02(x);
		tbl[2][i][3] = 0.5*k12(x);
	}
	
	/* remember that tables have been built */
	tabled = 1;
}



