%% demo_Prop_Th41pp_complex.m
%
% box algorithm (Theorem 4.1) using PSWFs
% - Fermi coordinates paraxial system
% - prolate expansion in parallel moving frame + coordinate transform
% - does not exploit spatial decay of wave packets / compression yet
% - optional zero-order approximation (coordinate transform without prolate expansion)
% - MEMORY INTENSIVE since total calculation domain is not fixed and all
%       box contributions are stored until the end (embarassing
%       box-wise parallelization)


clear all;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% PART I: INITIALIZATION  OF DEMO %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% -- add subdirectories
p=which('demo_Prop_Th41pp_complex'); index=strfind(p,'/'); p=p(1:index(end));addpath(genpath(p));

% --- initialize WP transform
N=256;                          % 2D data size
epsilon=1e-4;
rflag='complex';% curvelet transform options
% rflag='real';
conf=dpwave_conf(N,epsilon,rflag);
eer=epsilon;
Kmax=length(conf.N_nu);
pswf=load('pswf_1d_e4_c1_100.mat');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% --- PARAMETER SETUP

% -- operator
doAMPS=1;  % compute amplitudes
ZO=0;      % use zero order approximation instead of Prolate expansion

% -- scales and orientations to be computed
ANG=pi/180;
KK=4;
PSI=90;                          % ANGLE( xi1 , kx ) (degrees) of curvelet
DPSI=15;                         % max angle difference with neighboring scales to be treated

% -- physical parameters
Xmax=5;
Ymin=0.1;
Toff=8.0;
% Toff=9.7;
v=2;

% -- ray tracing and oversampling
DO_RT=1; dload=0;
Nxx=32; Ntt=Nxx;
logZP=1;    % x domain zero padding (power of two)
nt=400;

% -- velocity model
Xfact=4; Zfact=4;
X0=[0 25];
C0=2;
ST=3;
NOR=-0.4*C0;

% -- misc
verbose=3;
FIGNUM=2000;
if ZO; TSTR=['zero order approximation']; else; TSTR=['prolate approximation']; end;

Ymax=2*Xmax;
ZZ=(Ymax+Ymin)/2;
Nip=N; xy0=[0 0]; xy(1)=2*Xmax/(Nip-1); dxy=[2*Xmax, (Ymax-Ymin)]/(Nip);
OS=[1 0];

% -- create low velocity lens
Vxz.nx=ceil(Xfact*N); Vxz.nz=ceil(Zfact*N);
xv=linspace(-Xmax*Xfact,Xmax*Xfact,Vxz.nx); zv=linspace(0,Zfact*Ymax,Vxz.nz);
dx=diff(xv); Vxz.dx=dx(1); dz=diff(zv); Vxz.dz=dz(1); Vxz.fx=xv(1);Vxz.fz=zv(1);[XXX,ZZZ]=meshgrid(zv,xv); Vxz.x=XXX; Vxz.z=ZZZ;
Vxz.v=C0 + NOR*exp(-((XXX-X0(2)).^2+(ZZZ-X0(1)).^2)/ST^2);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% --- INPUT DATA as WP COEFFICIENTS
if strcmp(conf.cplx_flag,'complex');nudivide=1; else; nudivide=2; end; CBX=conf.boxgrid; clear mycoefs;
% initialize coeffs
mycoefs{1}{1}(1:CBX(1,1),1:CBX(1,2))=0;
for k=2:length(conf.N_nu); for l=1:conf.N_nu(k)/nudivide; mycoefs{k}{l}(1:CBX(k,1),1:CBX(k,2))=0; end; end
% input data: band-limited Dirac
for kid=1:length(KK)
    K=KK(kid);nu=find(conf.theta{K}>=PSI*ANG-eps);NU=nu(1);
    id1=find(conf.theta{K}/ANG>PSI-DPSI+0.01); id1=id1(1); id2=find(conf.theta{K}/ANG<PSI+DPSI-0.01); id2=id2(end);
    NUall=id1:id2;Nnu=length(NUall); p1=conf.boxgrid(K,1)/2+1; p2=conf.boxgrid(K,2)/2+1;
    for nid=1:Nnu; mycoefs{K}{NUall(nid)}(p1,p2)=1+1i; end
end
clear KK PSI NUall Nnu;
if verbose >1;myrec0=dpwave_recpp(mycoefs,conf); figure(FIGNUM+22); clf; subplot(211); imagesc([-1 1]*Xmax,[Ymin Ymax],real(myrec0)); colormap(bone); colorbar; title('real in'); subplot(212); imagesc([-1 1]*Xmax,[Ymin Ymax],imag(myrec0)); colormap(bone); colorbar; title('imag in');end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% MATCH TO PHYSICAL UNITS
Norm(1)=2*pi/(2*Xmax);
Norm(2)=2*pi/(Ymax-Ymin);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% <--- INITIALIZATIONS OF DEMO TERMINATED. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% PART II: ACTUAL COMPUTATION %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% FIND NONEMPTY SCALES AND ANGLES
KK=[];
for k=2:length(conf.N_nu)
    tmp_nu=[];
    for l=1:conf.N_nu(k)/nudivide
        tmp=find(mycoefs{k}{l}(:)~=0);
        if ~isempty(tmp); tmp_nu=[tmp_nu, l]; end
    end
    Boxes{k}=tmp_nu;
    if ~isempty(tmp_nu); KK=[KK, k]; end
end
% prepare for loop over boxes --> parallel
Nbox=0;
for kid=1:length(KK)
    K=KK(kid);NUall=Boxes{K};Nnu=length(NUall);
    for nuid=1:Nnu
        Nbox=Nbox+1;the_NU=NUall(nuid);kid_p(Nbox)=kid;nuid_p(Nbox)=nuid;the_NU_p(Nbox)=the_NU;K_p(Nbox)=K;
    end
end
N1=zeros(1,Nbox);N2=zeros(1,Nbox);NU=zeros(1,Nbox);idy1min=zeros(1,Nbox);idy2min=zeros(1,Nbox);idy1max=zeros(1,Nbox);idy2max=zeros(1,Nbox);y1_1=zeros(1,Nbox);y1_2=zeros(1,Nbox);y2_1=zeros(1,Nbox);y2_2=zeros(1,Nbox);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% LOOP NON-EMPTY SCALES AND ANGLES:
%% PROPAGATE

T_RT=0; T_BOX=0; T_RCT=0; runtime_1=clock;
% parfor boxid=1:Nbox %--> switch to this loop for parallel computation per box
for boxid=1:Nbox
    % -- this is the current box
    K=K_p(boxid);the_NU=the_NU_p(boxid);kid=kid_p(boxid);nuid=nuid_p(boxid); 
    runtime_1_bis=clock;
    if verbose; disp(['BOX K=',num2str(K),' -- theta=',num2str(round(100*conf.theta{K}(the_NU)*180/pi)/100)]); end
    [nu,xi,tt] = xi_k_nupp(conf,K, the_NU, Norm);     % geometry - normalized box frequencies & physical angle
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% --- ray tracing
    tic;
    DATe = DRT_mexProp_lensppcomplexIP(conf.theta{K}(the_NU),2^OS(1)*[-Xmax Xmax],[Ymin Ymax],Toff,Nxx,Ntt,[2^OS(1) 1]*Nip,xy0,dxy,Vxz,nt);
    T_RT=T_RT+toc;
    % -- read out and regularize d^2S/d(xi'')^2
    S2=DATe.d2S;S2(find(isnan(S2)))=0;S2(find(abs(S2)>100))=100*sign(S2(find(abs(S2)>100)));
    if doAMPS  % plane wave amplitudes
        amps=DATe.AMPip;
        for m=1:length(amps)
            ra=real(amps{m});ri=imag(amps{m});
            idr=find(abs(ra)>eps); idi=find(abs(ri)>eps);
            amps0=zeros(size(amps{m}));
            amps0(idr)=1.0*sign(real(amps{m}(idr)));
            amps0(idi)=amps0(idi)+1.0*1i*sign(imag(amps{m}(idi)));
            amps{m}=amps0;
        end
    else  % no amplitudes
        for m=1:length(amps); amps{m}=ones(size(DATe.AMPip{m})); end
    end
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% --- read coordinate transform
    XXcte=[];idXcte=[]; TTcte=[];idTcte=[]; YYcte=[];ZZcte=[];NXcte=[];vamps=[];
    for l=1:length(DATe.X)
        valid=find(DATe.X{l}>=-2^OS(1)*Xmax & DATe.X{l}<2^OS(1)*Xmax & DATe.T{l}>=Ymin & DATe.T{l} < Ymax);
        NXcte=[NXcte length(DATe.X{l}(valid))];
        XXcte=[XXcte; DATe.X{l}(valid)'];
        TTcte=[TTcte; DATe.T{l}(valid)'];
        idXcte=[idXcte; DATe.IND{l}(valid)'];
        AMPS{l}=[conj(amps{l}(valid)')];
        vamps=[vamps;AMPS{l}];
    end
    XM=2^OS(1)*Xmax; Xm=-XM;
    %% --- normalize to USFFT units
    xcte1=((XXcte-Xm)/(XM-Xm))*(2^OS(1)*N-1)-2^OS(1)*N/2;
    xcte2=((TTcte-Ymin)/(Ymax-Ymin))*(N-1)-N/2;
    XXcte=[]; TTcte=[]; valid=[];
    n1tmp=length(xcte1); n2tmp=length(xcte2); ntmp=min(n1tmp,n2tmp);
    xcte1=xcte1(1:ntmp); xcte2=xcte2(1:ntmp); NNcte=ntmp;
    yrange=DATe.yrange; zrange=DATe.zrange;
    T_RCT=T_RCT+toc;
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% --- expansion and coordinate transform
    tic;
    if ZO 
        [tmprec] = BoxZOppcomplex(mycoefs{K}{the_NU},conf.chi{K},xi,xcte1,xcte2,vamps,conf.N,Nip,[OS(1) 0],conf.epsilon); nE(boxid)=0;
    else
        [tmprec,nE(boxid)] = Box41ppcomplex(mycoefs{K}{the_NU},conf.chi{K},tt,xi,xcte1,xcte2,S2,vamps,pswf,conf.N,Nip,[OS(1) 0],Norm,conf.epsilon);
    end
    tt=[];xi=[];xcte1=[];xcte2=[];S2=[];AMPS=[]; xcte1=[]; xcte2=[];
    %% -- reshape to image
    imrse=zeros(DATe.N); idv=[1 cumsum(NXcte)];
    for k=1:length(NXcte);
        ctid=idv(k):idv(k+1); tmp=tmprec(ctid); imrse(idXcte(ctid))=imrse(idXcte(ctid))+tmp(:);
    end
    imrse=reshape(imrse,DATe.N(1),DATe.N(2))'; idXcte=[];tmprec=[];
    T_BOX=T_BOX+toc; runtime_2_bis=clock;
    %% -- save box contributions
    N1(boxid)=DATe.N(1);
    N2(boxid)=DATe.N(2);
    NU(boxid)=the_NU;
    idy1min(boxid)=DATe.ID0(1);
    idy2min(boxid)=DATe.ID0(2);
    idy1max(boxid)=DATe.ID0(1)+DATe.N(2)-1;
    idy2max(boxid)=DATe.ID0(2)+DATe.N(1)-1;
    y1_1(boxid)=yrange(1);
    y1_2(boxid)=yrange(2);
    y2_1(boxid)=zrange(1);
    y2_2(boxid)=zrange(2);
    PART{boxid}=imrse';
    imrse=[]; 
    
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% STACK UP BOX CONTRIBUTIONS TO FINAL IMAGE
for boxid=1:Nbox
    kid=kid_p(boxid);
    nuid=nuid_p(boxid);
    IMAGE.N1(kid,nuid)=N1(boxid);
    IMAGE.N2(kid,nuid)=N2(boxid);
    IMAGE.NU(kid,nuid)=NU(boxid);
    IMAGE.idy1min(kid,nuid)=idy1min(boxid);
    IMAGE.idy2min(kid,nuid)=idy2min(boxid);
    IMAGE.idy1max(kid,nuid)=idy1max(boxid);
    IMAGE.idy2max(kid,nuid)=idy2max(boxid);
    IMAGE.y1_1(kid,nuid)=y1_1(boxid);
    IMAGE.y1_2(kid,nuid)=y1_2(boxid);
    IMAGE.y2_1(kid,nuid)=y2_1(boxid);
    IMAGE.y2_2(kid,nuid)=y2_2(boxid);
    IMAGE.PART{kid}{nuid}=PART{boxid};
    IMAGE.nE(kid,nuid)=nE(boxid);
end
[IMAGE] = stick_imagepp(IMAGE);
runtime_2=clock;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% <--- PART II - ACTUAL COMPUTAION: DONE. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% -- display / plots
if ~strcmp('2007a',version('-release')); Ncore=matlabpool('size'); if ~Ncore; Ncore=1; end; else; Ncore=1; end
if verbose >1;
    figure(FIGNUM+20); clf; imagesc(IMAGE.fully1,IMAGE.fully2,real(IMAGE.FULL)'); colorbar; grid on; title([TSTR,' real']);colormap(bone);
    figure(FIGNUM+21); clf; imagesc(IMAGE.fully1,IMAGE.fully2,imag(IMAGE.FULL)'); colorbar; grid on; title([TSTR,' imag']);colormap(bone);
end
if verbose;
    if ~DO_RT; T_RT=0; end
    disp(' ');
    disp('------------------------------------------------');
    disp(['Runtime:                                ',num2str(ceil(100*etime(runtime_2,runtime_1))/100),'s']);
    disp(' ');
    disp(['Accumulative runtime  (',num2str(Ncore),' labs):         ',num2str(ceil(100*(T_RT+T_BOX))/100),'s']);
    if DO_RT; disp([' -- ray tracing / interpolation:        ',num2str(ceil(100*T_RT)/100),'s']); end
    if ZO;
        disp([' -- coordinate transform:               ',num2str(ceil(100*(T_BOX))/100),'s']);
    else
        disp([' -- prolate exp. / coordinate transf.:  ',num2str(ceil(100*(T_BOX))/100),'s']);
    end
    disp(' --------------------------------------'); disp(' ');
end
%%